#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2025, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2025. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #
import abc

from aps.beamline_driver.motion_management.facade import Movement, MotorsInfo, MotorsInfoLegacy
from aps.beamline_driver.beam_management.facade import Beam

from aps.common.data_structures import DictionaryWrapper

from aps.beamline_driver.beam_management.facade import IBeamManager
from aps.beamline_driver.motion_management.facade import IMotionManager, MethodCallingType

class AbstractOptimizedOpticsManager(IBeamManager, IMotionManager):

    def __init__(self, digital_twin_init_parameters: DictionaryWrapper, method_calling_type=MethodCallingType.STANDARD):
        self._digital_twin_init_parameters = digital_twin_init_parameters
        self._beam_manager: IBeamManager     = None
        self._motion_manager: IMotionManager = None
        self._initialize_digital_twin()

        self.set_method_calling_type(method_calling_type)

    def method_calling_type(self) -> MethodCallingType: return self.__method_calling_type

    def set_method_calling_type(self, value: MethodCallingType):
        self.__method_calling_type = value

        self._motion_manager.set_method_calling_type(self.__method_calling_type)
        self._beam_manager.set_method_calling_type(self.__method_calling_type)

        # DYNAMIC IMPLEMENTATION OF THE INTERFACE FOR BLUESKY/OPHYD
        if self.method_calling_type() == MethodCallingType.GENERATOR:
            # motion manager
            setattr(self, "move_motor",          self.__move_motor_generator)
            setattr(self, "get_motor_position",  self.__get_motor_position_generator)
            setattr(self, "move_motors",         self.__move_motors_generator)
            setattr(self, "get_motor_positions", self.__get_motor_positions_generator)
            # beam manager
            setattr(self, "set_initial_flux",    self.__set_initial_flux_generator)
            setattr(self, "get_beam",            self.__get_beam_generator)
        else:
            # motion manager
            setattr(self, "move_motor",          self.__move_motor)
            setattr(self, "get_motor_position",  self.__get_motor_position)
            setattr(self, "move_motors",         self.__move_motors)
            setattr(self, "get_motor_positions", self.__get_motor_positions)
            # beam manager
            setattr(self, "set_initial_flux",    self.__set_initial_flux)
            setattr(self, "get_beam",            self.__get_beam)

    @abc.abstractmethod
    def _initialize_digital_twin(self): raise NotImplementedError("here initialize self._beam_manager and self._motion_manager")

    # ------------------------------------------------------------------------
    # ------------------------------------------------------------------------
    # MOTION MANAGER
    # ------------------------------------------------------------------------
    # ------------------------------------------------------------------------

    def __move_motor_generator(self, motor_id: str, position: float, movement=Movement.ABSOLUTE):
        yield from self._motion_manager.move_motor(motor_id, position, movement=movement)

    def __get_motor_position_generator(self, motor_id: str) -> float:
        position = yield from self._motion_manager.get_motor_position(motor_id)
        return position

    def __move_motors_generator(self, motor_positions: dict, movement=Movement.ABSOLUTE):
        yield from self._motion_manager.move_motors(motor_positions, movement=movement)

    def __get_motor_positions_generator(self, motor_ids: list = None) -> dict:
        position = yield from self._motion_manager.get_motor_positions(motor_ids)
        return position

    def __move_motor(self, motor_id: str, position: float, movement=Movement.ABSOLUTE):
        self._motion_manager.move_motor(motor_id, position, movement=movement)

    def __get_motor_position(self, motor_id: str) -> float:
        return self._motion_manager.get_motor_position(motor_id)

    def __move_motors(self, motor_positions: dict, movement=Movement.ABSOLUTE):
        self._motion_manager.move_motors(motor_positions, movement=movement)

    def __get_motor_positions(self, motor_ids: list = None) -> dict:
        return self._motion_manager.get_motor_positions(motor_ids)

    def get_motors_info(self) -> MotorsInfo | MotorsInfoLegacy:
        return self._motion_manager.get_motors_info()

    # ------------------------------------------------------------------------
    # ------------------------------------------------------------------------
    # BEAM MANAGER
    # ------------------------------------------------------------------------
    # ------------------------------------------------------------------------

    def __set_initial_flux_generator(self, *args, **kwargs):
        yield from self._beam_manager.set_initial_flux(*args, **kwargs)

    def __set_initial_flux(self, *args, **kwargs):
        self._beam_manager.set_initial_flux(*args, **kwargs)

    def __get_beam_generator(self, *args, **kwargs) -> Beam:
        beam = yield from self._beam_manager.get_beam(*args, **kwargs)
        return beam

    def __get_beam(self, *args, **kwargs) -> Beam:
        return self._beam_manager.get_beam(*args, **kwargs)

    def get_beam_manager_id(self): return self._beam_manager.get_beam_manager_id()

    def get_beam_properties_type(self): return self._beam_manager.get_beam_properties_type()

    # ------------------------------------------------------------------------
    # ------------------------------------------------------------------------
    # COMMON
    # ------------------------------------------------------------------------
    # ------------------------------------------------------------------------

    def to_json(self, *args, **kwargs): self._motion_manager.to_json(*args, **kwargs)
    @classmethod
    def from_json(cls, file_name, *args, **kwargs): raise NotImplementedError("To be implemented")






#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2025, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2025. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #
import abc
import json
import warnings

import numpy
from aps.beamline_driver.common.legacy_keys import Keys
from aps.beamline_driver.facade import MethodCallingType
from aps.beamline_driver.motion_management.facade import IMotionManager, MotorsInfo, MotorsInfoLegacy, Movement, Units, MotorType, MotorInfo, MotorInfoLegacy

class AbstractMotionManager(IMotionManager):
    def __init__(self, motors_info: MotorsInfo | MotorsInfoLegacy =None, json_file: str=None, method_calling_type: MethodCallingType=MethodCallingType.STANDARD, platform=Keys.Platforms.DigitalTwin, **kwargs):
        if not motors_info is None:
            self.__motors_info = motors_info
        elif not json_file is None:
            self.__motors_info=AbstractMotionManager._load_motors_info_from_json(json_file)
            print(f"Motion Manager initialized from JSON file {json_file}")
        else:
            raise ValueError("Motion Manager cannot be initialized without motors info")

        self.set_method_calling_type(method_calling_type)
        self.__platform = platform

    def platform(self):
        return self.__platform

    def method_calling_type(self) -> MethodCallingType: return self.__method_calling_type

    def set_method_calling_type(self, value : MethodCallingType):
        self.__method_calling_type = value

        if self.method_calling_type() == MethodCallingType.GENERATOR:
            setattr(self, "_call_get_motor_position",  self.__call_get_motor_position_generator)
            setattr(self, "_call_move_motor",          self.__call_move_motor_generator)

            setattr(self, "move_motor",          self.__move_motor_generator)
            setattr(self, "get_motor_position",  self.__get_motor_position_generator)
            setattr(self, "move_motors",         self.__move_motors_generator)
            setattr(self, "get_motor_positions", self.__get_motor_positions_generator)
        else:

            setattr(self, "_call_get_motor_position",  self.__call_get_motor_position)
            setattr(self, "_call_move_motor",          self.__call_move_motor)

            setattr(self, "move_motor",          self.__move_motor)
            setattr(self, "get_motor_position",  self.__get_motor_position)
            setattr(self, "move_motors",         self.__move_motors)
            setattr(self, "get_motor_positions", self.__get_motor_positions)

    def _get_native_position(self, position: float, units: Units, motor_info: MotorInfo | MotorInfoLegacy):
        native_position = position

        if not units is None and (units != motor_info.units):
            if motor_info.type == MotorType.TRANSLATIONAL:
                if   units == Units.MICRON      and motor_info.units == Units.MILLIMETERS: native_position = position*1e-3
                elif units == Units.MILLIMETERS and motor_info.units == Units.MICRON:      native_position = position*1e3
            elif motor_info.type == MotorType.ROTATIONAL:
                if units == Units.DEGREES:
                    if motor_info.units   == Units.RADIANS:      native_position = numpy.radians(position)
                    elif motor_info.units == Units.MILLIRADIANS: native_position = numpy.radians(position)*1e3
                elif units.RADIANS:
                    if motor_info.units   == Units.DEGREES:      native_position = numpy.degrees(position)
                    elif motor_info.units == Units.MILLIRADIANS: native_position = position*1e3
                elif units.MILLIRADIANS:
                    if motor_info.units   == Units.RADIANS:      native_position = position*1e-3
                    elif motor_info.units == Units.DEGREES:      native_position = numpy.degrees(position*1e-3)
            else:
                warnings.warn(f"Requested units {units} are different from configured {motor_info.units} and no conversion is possible for this motor type {motor_info.type}")

        return motor_info.apply_resolution(native_position)

    def _check_absolute_position(self, native_position: float, current_position: float, movement : Movement, motor_info: MotorInfo | MotorInfoLegacy):
        absolute_position = native_position if movement == Movement.ABSOLUTE else native_position + current_position

        try:    boundaries = motor_info.boundaries_by_key(self.platform()) # legacy
        except: boundaries = motor_info.boundaries

        if absolute_position < boundaries[0] or absolute_position > boundaries[1]:
            raise ValueError("Requested position (" + str(native_position) + " is outside the boundaries (" + str(boundaries) + ")")

    def _convert_native_position(self, native_position: float, units: Units, motor_info: MotorInfo | MotorInfoLegacy):
        native_position = motor_info.apply_resolution(native_position)
        position        = native_position

        if not units is None and (units != motor_info.units):
            if motor_info.type == MotorType.TRANSLATIONAL:
                if   units == Units.MICRON      and motor_info.units == Units.MILLIMETERS: position = native_position*1e3
                elif units == Units.MILLIMETERS and motor_info.units == Units.MICRON:      position = native_position*1e-3
            elif motor_info.type == MotorType.ROTATIONAL:
                if units == Units.DEGREES:
                    if motor_info.units   == Units.RADIANS:      position = numpy.degrees(native_position)
                    elif motor_info.units == Units.MILLIRADIANS: position = numpy.degrees(native_position*1e-3)
                elif units.RADIANS:
                    if motor_info.units   == Units.DEGREES:      position = numpy.radians(native_position)
                    elif motor_info.units == Units.MILLIRADIANS: position = native_position*1e-3
                elif units.MILLIRADIANS:
                    if motor_info.units   == Units.RADIANS:      position = native_position*1e3
                    elif motor_info.units == Units.DEGREES:      position = numpy.radians(native_position)*1e3
            else:
                warnings.warn(f"Requested units {units} are different from configured {motor_info.units} and no conversion is possible for this motor type {motor_info.type}")

        return position

    # -----------------------------------------------------------------------
    # -----------------------------------------------------------------------
    # -----------------------------------------------------------------------
    # METHOD TYPE DYNAMIC MANAGEMENT -> PRIVATE METHODS
    # -----------------------------------------------------------------------
    # -----------------------------------------------------------------------
    # -----------------------------------------------------------------------

    def __call_get_motor_position(self, motor_id: str, units : Units = None, *args, **kwargs):
        return self._get_motor_position(motor_id, units=units, *args, **kwargs)

    def __call_get_motor_position_generator(self, motor_id: str, units : Units = None, *args, **kwargs):
        current_position = yield from self._get_motor_position(motor_id, units=units, *args, **kwargs)
        return current_position

    def __call_move_motor(self, motor_id: str, position: float, movement : Movement, units : Units = None, *args, **kwargs):
        self._move_motor(motor_id, position, movement, units, *args, **kwargs)

    def __call_move_motor_generator(self, motor_id: str, position: float, movement : Movement, units : Units = None, *args, **kwargs):
        yield from self._move_motor(motor_id, position, movement, units, *args, **kwargs)

    # -----------------------------------------------------------------------
    # -----------------------------------------------------------------------
    # -----------------------------------------------------------------------

    def __move_motor(self, motor_id: str, position: float, movement : Movement, units : Units = None, *args, **kwargs):
        motor_info = self.__motors_info.get_motor_info(motor_id)

        current_position = self._call_get_motor_position(motor_id, units=motor_info.units, *args, **kwargs)
        native_position  = self._get_native_position(position, units, motor_info)

        self._check_absolute_position(native_position, current_position, movement, motor_info)

        self._call_move_motor(motor_id, native_position, movement, units=motor_info.units, *args, **kwargs)

    # -----------------------------------------------------------------------

    def __get_motor_position(self, motor_id: str, units : Units = None, *args, **kwargs) -> float:
        motor_info      = self.__motors_info.get_motor_info(motor_id)

        native_position = self._call_get_motor_position(motor_id, units=motor_info.units, *args, **kwargs)

        return self._convert_native_position(native_position, units, motor_info)

    # -----------------------------------------------------------------------

    def __move_motor_generator(self, motor_id: str, position: float, movement : Movement, units : Units = None, *args, **kwargs):
        motor_info = self.__motors_info.get_motor_info(motor_id)

        current_position = yield from self._call_get_motor_position(motor_id, units=motor_info.units, *args, **kwargs)
        native_position  = self._get_native_position(position, units, motor_info)

        self._check_absolute_position(native_position, current_position, movement, motor_info)

        yield from self._call_move_motor(motor_id, native_position, movement, units=motor_info.units, *args, **kwargs)

    # -----------------------------------------------------------------------

    def __get_motor_position_generator(self, motor_id: str, units : Units = None, *args, **kwargs) -> float:
        motor_info      = self.__motors_info.get_motor_info(motor_id)

        native_position = yield from self._call_get_motor_position(motor_id, units=motor_info.units, *args, **kwargs)

        return self._convert_native_position(native_position, units, motor_info)

    # -----------------------------------------------------------------------

    def __move_motors_generator(self, motor_positions : dict, movement : Movement, units: dict=None, *args, **kwargs):
        for motor_id in motor_positions:
            yield from self.move_motor(motor_id, motor_positions[motor_id], movement, units.get(motor_id, None) if units is not None else None, *args, **kwargs)

    # -----------------------------------------------------------------------

    def __get_motor_positions_generator(self, motor_ids : list = None, units: dict=None, *args, **kwargs) -> dict:
        positions = {}
        for motor_id in (self.get_motors_info().get_motor_ids() if motor_ids is None else motor_ids):
            positions[motor_id] = yield from self.get_motor_position(motor_id, units.get(motor_id, None) if units is not None else None, *args, **kwargs)

        return positions

    # -----------------------------------------------------------------------

    def __move_motors(self, motor_positions : dict, movement : Movement, units: dict=None, *args, **kwargs):
        for motor_id in motor_positions: self.move_motor(motor_id, motor_positions[motor_id], movement, units.get(motor_id, None) if units is not None else None, *args, **kwargs)

    # -----------------------------------------------------------------------

    def __get_motor_positions(self, motor_ids : list = None, units: dict=None, *args, **kwargs) -> dict:
        return {motor_id : self.get_motor_position(motor_id, units.get(motor_id, None) if units is not None else None, *args, **kwargs) for motor_id in (self.get_motors_info().get_motor_ids() if motor_ids is None else motor_ids)}

    # -----------------------------------------------------------------------
    # -----------------------------------------------------------------------
    # -----------------------------------------------------------------------

    def get_motors_info(self) -> MotorsInfo | MotorsInfoLegacy: return self.__motors_info

    def to_json(self, file_name, *args, **kwargs):
        with open(file_name, "w") as file: json.dump(self.__motors_info.to_dictionary(), file, indent=4)
        print(f"Motion Manager saved to JSON file {file_name}")

    @classmethod
    def from_json(cls, file_name, *args, **kwargs): return cls(motors_info=cls._load_motors_info_from_json(file_name))

    @staticmethod
    def _load_motors_info_from_json(file_name) -> MotorsInfo | MotorsInfoLegacy:
        with open(file_name, "r") as file: dictionary = json.load(file)
        if MotorsInfoLegacy.is_legacy(dictionary): return MotorsInfoLegacy.from_dictionary(dictionary)
        else:                                      return MotorsInfo.from_dictionary(dictionary)

    # -----------------------------------------------------------------------

    @abc.abstractmethod
    def _move_motor(self, motor_id: str, position: float, movement : Movement, units : Units = None, *args, **kwargs): raise NotImplementedError
    @abc.abstractmethod
    def _get_motor_position(self, motor_id: str, units : Units = None, *args, **kwargs) -> float: raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2025, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2025. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #
import abc
import inspect
from typing import Type
import numpy as np
import os
from matplotlib.figure import Figure
from matplotlib.ticker import FuncFormatter
import cmasher as cmm

from aps.beamline_driver.facade import MethodCallingType
from aps.beamline_driver.beam_management.facade import BeamProperties, Beam, IBeamManager
from aps.common.data_structures import DictionaryWrapper

class AbstractBeamManager(IBeamManager):
    def __init__(self, method_calling_type: MethodCallingType=MethodCallingType.STANDARD, **kwargs):
        self.__input_parameters = DictionaryWrapper(**kwargs)

        self._initial_flux      = None
        self.set_method_calling_type(method_calling_type)

    def method_calling_type(self): return self.__method_calling_type

    def set_method_calling_type(self, value):
        self.__method_calling_type = value

        # DYNAMIC IMPLEMENTATION OF THE INTERFACE FOR BLUESKY/OPHYD
        if self.method_calling_type() == MethodCallingType.STANDARD:
            setattr(self, "_call_set_initial_flux", self.__call_set_initial_flux)
            setattr(self, "_call_get_beam", self.__call_get_beam)

            setattr(self, "set_initial_flux", self.__set_initial_flux)
            setattr(self, "get_beam", self.__get_beam)
        elif self.method_calling_type() == MethodCallingType.GENERATOR:
            setattr(self, "_call_set_initial_flux",  self.__call_set_initial_flux_generator)
            setattr(self, "_call_get_beam",          self.__call_get_beam_generator)

            setattr(self, "set_initial_flux",    self.__set_initial_flux_generator)
            setattr(self, "get_beam",            self.__get_beam_generator)

    # -----------------------------------------------------------------------
    # -----------------------------------------------------------------------
    # -----------------------------------------------------------------------
    # METHOD TYPE DYNAMIC MANAGEMENT -> PRIVATE METHODS
    # -----------------------------------------------------------------------
    # -----------------------------------------------------------------------
    # -----------------------------------------------------------------------

    def __call_set_initial_flux(self, *args, **kwargs):
        self._set_initial_flux(*args, **kwargs)

    def __call_set_initial_flux_generator(self, *args, **kwargs):
        yield from self._set_initial_flux(*args, **kwargs)

    def __call_get_beam(self, *args, **kwargs):
        return self._get_beam(*args, **kwargs)

    def __call_get_beam_generator(self, *args, **kwargs):
        beam = yield from self._get_beam(*args, **kwargs)
        return beam

    def __set_initial_flux(self, *args, **kwargs):
        self._call_set_initial_flux(*args, **kwargs)

    def __set_initial_flux_generator(self, *args, **kwargs):
        yield from self._call_set_initial_flux(*args, **kwargs)

    def __get_beam(self, *args, **kwargs):
        return self._call_get_beam(*args, **kwargs)

    def __get_beam_generator(self, *args, **kwargs):
        beam = yield from self._call_get_beam(*args, **kwargs)
        return beam

    @abc.abstractmethod
    def _set_initial_flux(self, *args, **kwargs): raise NotImplementedError

    @abc.abstractmethod
    def _get_beam(self, *args, **kwargs) -> Beam: raise NotImplementedError

    def get_beam_manager_id(self): return self.__input_parameters.get_parameter("beam_manager_id", None)

    def _get_input_parameters(self) -> DictionaryWrapper: return self.__input_parameters

    def _plot_beam(self, beam_properties, **kwargs):
        if kwargs.get("save_image", False) or kwargs.get("show_beam", False):
            beam_histogram = beam_properties.get_parameter(self.get_beam_histogram_name())

            if not kwargs.get("flip", False):
                hh = beam_histogram.get_parameter("coordinate_h")
                vv = beam_histogram.get_parameter("coordinate_v")
                data_2D = beam_histogram.get_parameter("histogram")
            else:
                hh = beam_histogram.get_parameter("coordinate_v") # Z is horizontal
                vv = beam_histogram.get_parameter("coordinate_h") # X is horizontal
                data_2D = beam_histogram.get_parameter("histogram").T

            xrange = kwargs.get("xrange", None)
            yrange = kwargs.get("yrange", None)

            if xrange is None: xrange = [np.min(hh), np.max(hh)]
            if yrange is None: yrange = [np.min(vv), np.max(vv)]

            # Define a custom formatter function
            def custom_formatter(x, pos): return f'{x:.2f}'

            plt = kwargs.get("plt", None)

            if plt is None: fig = Figure(figsize=[9, 8], constrained_layout=True)
            else:           fig = plt.figure(figsize=[9, 8], constrained_layout=True)
            ax = fig.gca()

            image = fig.gca().pcolormesh(hh, vv, data_2D.T, cmap=cmm.sunburst_r, rasterized=True)
            fig.gca().set_xlim(xrange[0], xrange[1])
            fig.gca().set_ylim(yrange[0], yrange[1])

            # Generate 10 equally spaced ticks within the specified range
            x_ticks = np.linspace(xrange[0], xrange[1], 11)
            y_ticks = np.linspace(yrange[0], yrange[1], 11)

            # Set the ticks on the axes
            fig.gca().set_xticks(x_ticks)
            fig.gca().set_yticks(y_ticks)

            # Apply the custom formatter to the axes
            fig.gca().xaxis.set_major_formatter(FuncFormatter(custom_formatter))
            fig.gca().yaxis.set_major_formatter(FuncFormatter(custom_formatter))

            fig.gca().axhline(0, color="gray", ls="--", linewidth=1, alpha=0.7)
            fig.gca().axvline(0, color="gray", ls="--", linewidth=1, alpha=0.7)
            fig.gca().set_xlabel("Horizontal ($\\mu$m)")
            fig.gca().set_ylabel("Vertical ($\\mu$m)")

            cbar = fig.colorbar(mappable=image, pad=0.01, aspect=30, shrink=0.6)

            ax.set_aspect("equal")

            labels = self.get_labels()

            text = (rf"{'fwhm(H)':<6}  = {beam_properties.get_parameter(labels[0]) : 3.1f} $\mu$m" + "\n" +
                    rf"{'fwhm(V)':<6}  = {beam_properties.get_parameter(labels[1]) : 3.1f} $\mu$m" + "\n" +
                    rf"{'sigma(H)':<6} = {beam_properties.get_parameter(labels[2]) : 3.1f} $\mu$m" + "\n" +
                    rf"{'sigma(V)':<6} = {beam_properties.get_parameter(labels[3]) : 3.1f} $\mu$m" + "\n" +
                    rf"{'peak(H)':<6}  = {beam_properties.get_parameter(labels[4]) : 3.1f} $\mu$m" + "\n" +
                    rf"{'peak(V)':<6}  = {beam_properties.get_parameter(labels[5]) : 3.1f} $\mu$m" + "\n" +
                    rf"{'pI':<10}      = {beam_properties.get_parameter(labels[6]) : 1.3e} ph")
            ax.text(0.75, 0.7, text, color="black", alpha=0.9, fontsize=9, bbox=dict(facecolor="white", edgecolor="gray", alpha=0.7), transform=ax.transAxes)

            trial_num = kwargs.get("trial_num", None)
            if not trial_num is None: fig.gca().text(0.02, 0.95, f"Trial #{trial_num}", fontsize=10, color="blue", transform=ax.transAxes)

            cbar.ax.text(0.5, 1.05, "pI", transform=cbar.ax.transAxes, ha="center", va="bottom", fontsize=10, color="black")

            if kwargs.get("save_image", False):
                file_name = kwargs.get("file_name", None)
                if not file_name is None:
                    fig.savefig(os.path.join(kwargs.get("directory_name", os.path.abspath(os.curdir)), file_name))
                    fig.canvas.draw_idle()

            if kwargs.get("show_beam", False):
                plt = kwargs.get("plt", fig)
                plt.show()

    @classmethod
    @abc.abstractmethod
    def get_beam_histogram_name(cls):  raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def get_labels(cls): raise  NotImplementedError


class DirectBeamImageProperties:
    CENTROID_H         = "centroid-h"
    CENTROID_V         = "centroid-v"
    PEAK_LOCATION_H    = "peak-distance-h"
    PEAK_LOCATION_V    = "peak-distance-v"
    FWHM_H             = "fwhm-h"
    FWHM_V             = "fwhm-v"
    SIGMA_H            = "sigma-h"
    SIGMA_V            = "sigma-v"
    PEAK_INTENSITY     = "peak-intensity"
    INTEGRAL_INTENSITY = "integral-intensity"

    __list = [CENTROID_H, CENTROID_V, PEAK_LOCATION_H, PEAK_LOCATION_V, FWHM_H, FWHM_V, SIGMA_H, SIGMA_V, PEAK_INTENSITY, INTEGRAL_INTENSITY]

    @classmethod
    def to_list(cls): return cls.__list

class WavefrontAnalysisProperties:
    INTENSITY_CENTROID_H             = "intensity-centroid-h"
    INTENSITY_CENTROID_V             = "intensity-centroid-v"
    INTENSITY_FWHM_H                 = "intensity-fwhm-h"
    INTENSITY_FWHM_V                 = "intensity-fwhm-v"
    INTENSITY_SIGMA_H                = "intensity-sigma-h"
    INTENSITY_SIGMA_V                = "intensity-sigma-v"
    INTENSITY_MAXIMUM                = "intensity-max"
    INTENSITY_INTEGRAL               = "intensity-integral"
    FOCUS_Z_POSITION_H               = "focus-z-position-h"
    FOCUS_Z_POSITION_V               = "focus-z-position-v"

    __list = [INTENSITY_CENTROID_H,
              INTENSITY_CENTROID_V,
              INTENSITY_FWHM_H,
              INTENSITY_FWHM_V,
              INTENSITY_SIGMA_H,
              INTENSITY_SIGMA_V,
              INTENSITY_MAXIMUM,
              INTENSITY_INTEGRAL,
              FOCUS_Z_POSITION_H,
              FOCUS_Z_POSITION_V]

    @classmethod
    def to_list(cls): return cls.__list

class SpeckleAnalysisProperties:
    FWHM_H          = "fwhm-h"
    FWHM_V          = "fwhm-v"
    SPECKLE_SIZE_H  = "speckle-size-h"
    SPECKLE_SIZE_V  = "speckle-size-v"

    __list = [FWHM_H, FWHM_V, SPECKLE_SIZE_H, SPECKLE_SIZE_V]

    @classmethod
    def to_list(cls): return cls.__list


class DirectBeamImageBeamManager(AbstractBeamManager):
    def __init__(self, method_calling_type: MethodCallingType=MethodCallingType.STANDARD, **kwargs):
        super(DirectBeamImageBeamManager, self).__init__(method_calling_type=method_calling_type, **kwargs)

    def get_beam_properties_type(self) -> Type: return DirectBeamImageProperties

    def get_beam_properties(self, beam : Beam, **kwargs) -> BeamProperties:
        beam_histogram = self._get_beam_histogram(beam, **kwargs)

        centroid_h, centroid_v                 = _get_centroids(beam_histogram)
        peak_distance_h, peak_distance_v, _, _ = _get_peak_coordinates(beam_histogram)
        fwhm_h, fwhm_v                         = _get_fwhms(beam_histogram)
        sigma_h, sigma_v                       = _get_sigmas(beam_histogram)

        beam_properties = BeamProperties()
        beam_properties.set_parameter(DirectBeamImageProperties.CENTROID_H, centroid_h)
        beam_properties.set_parameter(DirectBeamImageProperties.CENTROID_V, centroid_v)
        beam_properties.set_parameter(DirectBeamImageProperties.PEAK_LOCATION_H, peak_distance_h)
        beam_properties.set_parameter(DirectBeamImageProperties.PEAK_LOCATION_V, peak_distance_v)
        beam_properties.set_parameter(DirectBeamImageProperties.FWHM_H, fwhm_h)
        beam_properties.set_parameter(DirectBeamImageProperties.FWHM_V, fwhm_v)
        beam_properties.set_parameter(DirectBeamImageProperties.SIGMA_H, sigma_h)
        beam_properties.set_parameter(DirectBeamImageProperties.SIGMA_V, sigma_v)
        beam_properties.set_parameter(DirectBeamImageProperties.PEAK_INTENSITY, _get_peak_intensity(beam_histogram))

        integral_instensity = _get_integral_intensity(beam_histogram)

        beam_properties.set_parameter(DirectBeamImageProperties.INTEGRAL_INTENSITY, integral_instensity)
        beam_properties.set_parameter("is_empty_beam", integral_instensity <= self._get_input_parameters().get_parameter("empty_beam_threshold", 0.0))
        beam_properties.set_parameter(self.get_beam_histogram_name(), beam_histogram)

        self._plot_beam(beam_properties, **kwargs)

        return beam_properties

    @abc.abstractmethod
    def _get_beam_histogram(self, beam : Beam, **kwargs) -> BeamProperties: raise NotImplementedError

    @classmethod
    def get_beam_histogram_name(cls): return "beam_histogram"

    @classmethod
    def get_labels(cls): return [DirectBeamImageProperties.FWHM_H,
                                 DirectBeamImageProperties.FWHM_V,
                                 DirectBeamImageProperties.SIGMA_H,
                                 DirectBeamImageProperties.SIGMA_V,
                                 DirectBeamImageProperties.PEAK_LOCATION_H,
                                 DirectBeamImageProperties.PEAK_LOCATION_V,
                                 DirectBeamImageProperties.PEAK_INTENSITY]

class WavefrontAnalysisBeamManager(AbstractBeamManager):
    def __init__(self, method_calling_type: MethodCallingType = MethodCallingType.STANDARD, **kwargs):
        super(WavefrontAnalysisBeamManager, self).__init__(method_calling_type=method_calling_type, **kwargs)

    def get_beam_properties_type(self) -> Type: return WavefrontAnalysisProperties

    def get_beam_properties(self, beam : Beam, **kwargs) -> BeamProperties:
        intensity_histogram, phase_histogram = self._get_wavefront_histogram(beam, **kwargs)

        centroid_h, centroid_v                 = _get_centroids(intensity_histogram)
        peak_distance_h, peak_distance_v, _, _ = _get_peak_coordinates(intensity_histogram)
        fwhm_h, fwhm_v                         = _get_fwhms(intensity_histogram)
        sigma_h, sigma_v                       = _get_sigmas(intensity_histogram)

        beam_properties = BeamProperties()
        beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_CENTROID_H, centroid_h)
        beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_CENTROID_V, centroid_v)
        beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_FWHM_H, fwhm_h)
        beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_FWHM_V, fwhm_v)
        beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_SIGMA_H, sigma_h)
        beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_SIGMA_V, sigma_v)
        beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_MAXIMUM, _get_peak_intensity(intensity_histogram))

        integral_intensity = _get_integral_intensity(intensity_histogram)

        beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_INTEGRAL, integral_intensity)
        beam_properties.set_parameter("is_empty_beam", integral_intensity <= self._get_input_parameters().get_parameter("empty_beam_threshold", 0.0))
        beam_properties.set_parameter(self.get_beam_histogram_name(), intensity_histogram)

        beam_properties.set_parameter(WavefrontAnalysisProperties.FOCUS_Z_POSITION_H, phase_histogram.get_parameter("focus_z_position_h"))
        beam_properties.set_parameter(WavefrontAnalysisProperties.FOCUS_Z_POSITION_V, phase_histogram.get_parameter("focus_z_position_v"))

        self._plot_beam(beam_properties, **kwargs)

        return beam_properties

    @abc.abstractmethod
    def _get_wavefront_histogram(self, beam : Beam, **kwargs) -> BeamProperties: raise NotImplementedError

    @classmethod
    def get_beam_histogram_name(cls): return "intensity_histogram"

    @classmethod
    def get_labels(cls): return [WavefrontAnalysisProperties.INTENSITY_FWHM_H,
                                 WavefrontAnalysisProperties.INTENSITY_FWHM_V,
                                 WavefrontAnalysisProperties.INTENSITY_SIGMA_H,
                                 WavefrontAnalysisProperties.INTENSITY_SIGMA_V,
                                 WavefrontAnalysisProperties.FOCUS_Z_POSITION_H,
                                 WavefrontAnalysisProperties.FOCUS_Z_POSITION_V,
                                 WavefrontAnalysisProperties.INTENSITY_MAXIMUM]


class SpeckleAnalysisBeamManager(AbstractBeamManager):
    def __init__(self, method_calling_type: MethodCallingType=MethodCallingType.STANDARD, **kwargs):
        super(SpeckleAnalysisBeamManager, self).__init__(method_calling_type=method_calling_type, **kwargs)

    def get_beam_properties_type(self) -> Type: return SpeckleAnalysisProperties

    def get_beam_properties(self, beam : Beam, **kwargs) -> BeamProperties:
        speckle_data = self._extract_speckle_data(beam, **kwargs)

        beam_properties = BeamProperties()
        beam_properties.set_parameter("raw_data", speckle_data["raw_data"])
        beam_properties.set_parameter(SpeckleAnalysisProperties.FWHM_H, speckle_data[SpeckleAnalysisProperties.FWHM_H])
        beam_properties.set_parameter(SpeckleAnalysisProperties.FWHM_V, speckle_data[SpeckleAnalysisProperties.FWHM_V])
        beam_properties.set_parameter(SpeckleAnalysisProperties.SPECKLE_SIZE_H, speckle_data[SpeckleAnalysisProperties.SPECKLE_SIZE_H])
        beam_properties.set_parameter(SpeckleAnalysisProperties.SPECKLE_SIZE_V, speckle_data[SpeckleAnalysisProperties.SPECKLE_SIZE_V])
        beam_properties.set_parameter(self.get_beam_histogram_name() + "_h", speckle_data[self.get_beam_histogram_name() + "_h"])
        beam_properties.set_parameter(self.get_beam_histogram_name() + "_v", speckle_data[self.get_beam_histogram_name() + "_v"])

        beam_properties.set_parameter("coordinate_h", speckle_data["coordinate_h"])
        beam_properties.set_parameter("coordinate_v", speckle_data["coordinate_v"])

        self._plot_beam(beam_properties, **kwargs)

        return beam_properties

    @abc.abstractmethod
    def _extract_speckle_data(self, beam : Beam, **kwargs) -> dict: raise NotImplementedError

    @classmethod
    def  get_beam_histogram_name(cls): return "speckle_data"

    @classmethod
    def get_labels(cls): return SpeckleAnalysisProperties.to_list()



from aps.common.plot.image import get_average, get_peak_location_2D, get_fwhm, get_sigma

def _get_centroids(beam_histogram: BeamProperties):
    return get_average(beam_histogram.get_parameter("histogram_h"), beam_histogram.get_parameter("coordinate_h")), \
           get_average(beam_histogram.get_parameter("histogram_v"), beam_histogram.get_parameter("coordinate_v"))

def _get_peak_coordinates(beam_histogram: BeamProperties):
    return get_peak_location_2D(beam_histogram.get_parameter("coordinate_h"),
                                beam_histogram.get_parameter("coordinate_v"),
                                beam_histogram.get_parameter("histogram"))

def _get_fwhms(beam_histogram: BeamProperties):
    fwhm_h, _, _ = get_fwhm(beam_histogram.get_parameter("histogram_h"), beam_histogram.get_parameter("coordinate_h"))
    fwhm_v, _, _ = get_fwhm(beam_histogram.get_parameter("histogram_v"), beam_histogram.get_parameter("coordinate_v"))

    return fwhm_h, fwhm_v

def _get_sigmas(beam_histogram: BeamProperties):
    return get_sigma(beam_histogram.get_parameter("histogram_h"), beam_histogram.get_parameter("coordinate_h")), \
           get_sigma(beam_histogram.get_parameter("histogram_v"), beam_histogram.get_parameter("coordinate_v"))

def _get_peak_intensity(beam_histogram: BeamProperties) -> float:
    _, _, i, j = _get_peak_coordinates(beam_histogram)
    histogram   = beam_histogram.get_parameter("histogram")
    shape       = histogram.shape

    return np.average(histogram[max(0, i-1): min(shape[0], i+1), max(0, j-1) : min(shape[1], j+1)])

def _get_integral_intensity(beam_histogram: BeamProperties) -> float:
    return beam_histogram.get_parameter("histogram").sum()





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2025, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2025. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #

from aps.beamline_driver.motion_management.facade import MotorsInfo, MotorsInfoLegacy, Movement, Units, MethodCallingType
from aps.beamline_driver.motion_management.abstract_motion_manager import AbstractMotionManager, Keys

from epics import PV

class EpicsMotionManager(AbstractMotionManager):

    def __init__(self, motors_info: MotorsInfo | MotorsInfoLegacy=None, json_file: str=None, **kwargs):
        super(EpicsMotionManager, self).__init__(motors_info=motors_info,
                                                 json_file=json_file,
                                                 method_calling_type=MethodCallingType.STANDARD,
                                                 platform=Keys.Platforms.Beamline)
        self._initialize_pvs(**kwargs)

    def _initialize_pvs(self, **kwargs):
        motors_info = self.get_motors_info()

        self._pvs_put = {}
        self._pvs_get = {}
        for motor_id in motors_info.get_motor_ids():
            motor_info = motors_info.get_motor_info(motor_id)
            self._pvs_put[motor_id] = PV(motor_info.legacy_ids[0])
            self._pvs_get[motor_id] = PV(motor_info.legacy_ids[1])

    def _move_motor(self, motor_id: str, position: float, movement : Movement, units : Units = None, *args, **kwargs):
        native_position = self._get_native_position(position,
                                                    units,
                                                    self.get_motors_info().get_motor_info(motor_id)) # redundant, but harmless

        self._move_epics_motor(motor_id, native_position, movement, wait=kwargs.get("wait", True))

    def _get_motor_position(self, motor_id: str, units : Units = None, *args, **kwargs) -> float:
        native_position = self._get_epics_motor_native_position(motor_id)

        return self._convert_native_position(native_position,
                                             units,
                                             self.get_motors_info().get_motor_info(motor_id)) # redundant, but harmless

    def _move_epics_motor(self, motor_id, native_position: float, movement : Movement, wait=True):
        if movement == Movement.ABSOLUTE:   self._pvs_put[motor_id].put(native_position, wait=wait)
        elif movement == Movement.RELATIVE: self._pvs_put[motor_id].put(self._pvs_get[motor_id].get() + native_position, wait=wait)
        else: raise ValueError("Movement not recognized")

    def _get_epics_motor_native_position(self, motor_id):
        return self._pvs_get[motor_id].get()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2025, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2025. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #

import numpy as np
from aps.beamline_driver.beam_management.facade import BeamProperties, Beam, MethodCallingType
from aps.beamline_driver.beam_management.abstract_beam_manager import WavefrontAnalysisBeamManager

class SRWBeamManager(WavefrontAnalysisBeamManager):
    def __init__(self, method_calling_type=MethodCallingType.STANDARD, **kwargs):
        super(SRWBeamManager, self).__init__(method_calling_type=method_calling_type, **kwargs)

        self._to_micron = kwargs.get("to_micron", True)
        self._xrange    = kwargs.get("xrange", None)
        self._yrange    = kwargs.get("yrange", None)
        if not self._xrange is None: self._xrange = (np.array(self._xrange) * (1e-6 if self._to_micron else 1.0)).tolist()
        if not self._yrange is None: self._yrange = (np.array(self._yrange) * (1e-6 if self._to_micron else 1.0)).tolist()

    def _get_wavefront_histogram(self, beam : Beam, **kwargs) -> BeamProperties:
        srw_wavefront = beam.data

        intensity_histogram = self._get_wavefront_intensity(srw_wavefront, self._xrange, self._yrange, self._to_micron)

        # ------------------------------------------------------
        # FOR FUTURE APPLICATIONS
        #
        # e, h, v, p = self.output_wavefront.get_phase()
        # ticket_p              = self._get_ticket_2D(h, v, p[int(e.size / 2)])
        # ticket_p['histogram'] = unwrap(ticket_p['histogram'])
        # ------------------------------------------------------

        phase_histogram = BeamProperties(phase_avg_radius_h=srw_wavefront.Rx, phase_avg_radius_v=srw_wavefront.Ry)

        return intensity_histogram, phase_histogram

    @classmethod
    def _get_wavefront_intensity(cls, srw_wavefront, xrange, yrange, to_micron):
        e, h, v, i = srw_wavefront.get_intensity(multi_electron=False) # Single Electron -> provides the wavefront

        ticket_i = cls._get_ticket_2D(h, v, i[int(e.size / 2)])

        coordinate_h = ticket_i['bin_h']
        coordinate_v = ticket_i['bin_v']

        if not xrange is None: range_x = np.where(np.logical_and(coordinate_h >= xrange[0], coordinate_h <= xrange[1]))
        else:                  range_x = np.where(np.logical_and(coordinate_h >= np.min(coordinate_h), coordinate_h <= np.max(coordinate_h)))
        if not yrange is None: range_y = np.where(np.logical_and(coordinate_v >= yrange[0], coordinate_v <= yrange[1]))
        else:                  range_y = np.where(np.logical_and(coordinate_v >= np.min(coordinate_v), coordinate_v <= np.max(coordinate_v)))

        coordinate_h = ticket_i['bin_h'][range_x] * (1e6 if to_micron else 1.0)
        coordinate_v = ticket_i['bin_v'][range_y] * (1e6 if to_micron else 1.0)

        nbins_h = len(coordinate_h)
        nbins_v = len(coordinate_v)

        histogram = np.zeros((nbins_h, nbins_v))
        for row, i in zip(ticket_i['histogram'][range_x], range(nbins_h)): histogram[i, :] = row[range_y]

        histogram_h = np.sum(histogram, axis=1)
        histogram_v = np.sum(histogram, axis=0)

        return BeamProperties(histogram=histogram,
                              coordinate_h=coordinate_h,
                              coordinate_v=coordinate_v,
                              histogram_h=histogram_h,
                              histogram_v=histogram_v)

    @classmethod
    def _get_ticket_2D(cls, x_array, y_array, z_array):
        ticket = {'error':0}
        ticket['nbins_h'] = len(x_array)
        ticket['nbins_v'] = len(y_array)

        xrange = [x_array.min(), x_array.max() ]
        yrange = [y_array.min(), y_array.max() ]

        hh = z_array
        hh_h = hh.sum(axis=1)
        hh_v = hh.sum(axis=0)
        xx = x_array
        yy = y_array

        ticket['xrange'] = xrange
        ticket['yrange'] = yrange
        ticket['bin_h'] = xx
        ticket['bin_v'] = yy
        ticket['histogram'] = hh
        ticket['histogram_h'] = hh_h
        ticket['histogram_v'] = hh_v

        return ticket



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2025, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2025. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #

from PIL import Image
import numpy
import os

PIXEL_SIZE = 6.5e-7
IMAGE_SIZE_PIXEL_HxV = [2160, 2560]

def load_image(file_path):
    if os.path.exists(file_path): return numpy.array(numpy.array(Image.open(file_path))).astype(numpy.float32).T
    else:                         raise ValueError('Error: wrong data path. No data is loaded:' + file_path)

image = load_image(os.path.join(os.path.abspath(os.curdir), "../../../../../../../../../../../../Library/CloudStorage/Box-Box/Luca-Runyu/AI-Controller/tiff_noise/sample_300ms_00001.tif"))

factor = 1e6

h_coord = numpy.linspace(-IMAGE_SIZE_PIXEL_HxV[0] / 2, IMAGE_SIZE_PIXEL_HxV[0] / 2, IMAGE_SIZE_PIXEL_HxV[0]) * PIXEL_SIZE * factor
v_coord = numpy.linspace(-IMAGE_SIZE_PIXEL_HxV[1] / 2, IMAGE_SIZE_PIXEL_HxV[1] / 2, IMAGE_SIZE_PIXEL_HxV[1]) * PIXEL_SIZE * factor

from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm

dx = (h_coord[1]-h_coord[0])/2.
dy = (v_coord[1]-v_coord[0])/2.
extent = [h_coord[0]-dx, h_coord[-1]+dx, v_coord[0]-dy, v_coord[-1]+dy]

plt.figure(1)
plt.imshow(image.T, extent=extent, norm=LogNorm(vmin=0.01, vmax=numpy.max(image)))
plt.title("Image with noise (log)")
plt.xlabel("Horizontal [um]")
plt.ylabel("Vertical [um]")
plt.colorbar()

_, (ax1, ax2) = plt.subplots(1, 2)
plt.title("Projections (H - V, log)")
ax1.set_yscale('log')
ax2.set_yscale('log')
ax1.set_xlabel("Horizontal [um]")
ax2.set_xlabel("Vertical [um]")
ax1.plot(h_coord, image.sum(axis=1))
ax2.plot(v_coord, image.sum(axis=0))

from aps.beamline_driver.common.utility import calculate_projections_over_noise

image, histogram_h, histogram_v = calculate_projections_over_noise(histogram=image,
                                                                   noise_max_pos=10,
                                                                   noise_threshold=1.1)

plt.figure(3)
plt.imshow(image.T, extent=extent, norm=LogNorm(vmin=0.01, vmax=numpy.max(image)))
plt.title("Image denoised (log)")
plt.xlabel("Horizontal [um]")
plt.ylabel("Vertical [um]")
plt.colorbar()

_, (ax1, ax2) = plt.subplots(1, 2)
plt.title("Projections (H - V, log) denoised")
ax1.set_yscale('log')
ax2.set_yscale('log')
ax1.set_xlabel("Horizontal [um]")
ax2.set_xlabel("Vertical [um]")
ax1.plot(h_coord, histogram_h)
ax2.plot(v_coord, histogram_v)

plt.show()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2025, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2025. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #
import os
import numpy as np
from scipy.ndimage.measurements import center_of_mass

from aps.beamline_driver.beam_management.facade import Beam, BeamProperties, MethodCallingType
from aps.beamline_driver.beam_management.abstract_beam_manager import (DirectBeamImageBeamManager,
                                                                                   DirectBeamImageProperties)
from aps.beamline_driver.common.utility import calculate_projections_over_noise

from aps.common.driver.beamline.generic_camera import GenericCamera, CameraInitializationFile

class EpicsCameraDirectBeamManager(DirectBeamImageBeamManager):
    def __init__(self, camera : GenericCamera = None, measurement_directory=None, **kwargs):
        super(EpicsCameraDirectBeamManager, self).__init__(method_calling_type=MethodCallingType.STANDARD, **kwargs)

        if not camera is None:
            self.__camera = camera
        else:
            if not CameraInitializationFile.is_initialized():
                CameraInitializationFile.initialize()
                CameraInitializationFile.store()

            self.__camera = GenericCamera(measurement_directory=(measurement_directory if not measurement_directory is None else os.path.abspath(os.curdir)))

    def get_beam(self, *args, **kwargs) -> Beam:
        try:    self.__camera.save_status() # collect beamline original setups
        except: pass

        image_index  = kwargs.get("image_index", 1)
        index_digits = kwargs.get("index_digits", 4)
        units        = kwargs.get("units", "mm")
        use_stream   = kwargs.get("use_stream", False)

        try:
            self.__camera.collect_single_shot_image(index=image_index)

            if use_stream: image, coordinate_h, coordinate_v = self.__camera.get_image_stream_data(units=units)
            else:          image, coordinate_h, coordinate_v = self.__camera.get_image_data(image_index=image_index, index_digits=index_digits, units=units)

            beam_data = {}
            beam_data["coordinate_h"] = coordinate_h
            beam_data["coordinate_v"] = coordinate_v
            beam_data["image"]        = image

            try:    self.__camera.end_collection()
            except: pass
            try:    self.__camera.restore_status() # restore beamline original setups
            except: pass

            return Beam(beam_data=beam_data)
        except Exception as e:
            try:    self.__camera.end_collection()
            except: pass
            try:    self.__camera.restore_status() # restore beamline original setups
            except: pass

            raise e

    def get_hardware_device(self): return self.__camera

    def get_beam_properties(self, beam : Beam, **kwargs) -> BeamProperties:
        auto_crop_defocused_image        = kwargs.get("auto_crop_defocused_image", False)

        if not auto_crop_defocused_image:
            return super(EpicsCameraDirectBeamManager, self).get_beam_properties(beam=beam, **kwargs)
        else:
            crop_strip_width = kwargs.get("crop_strip_width", 50)
            crop_threshold   = kwargs.get("crop_threshold", None)

            coordinate_h = beam.data["coordinate_h"]
            coordinate_v = beam.data["coordinate_v"]
            image        = beam.data["image"]

            if crop_threshold is None: crop_threshold = np.average(image)

            footprint = np.ones(image.shape) * (image > crop_threshold)

            center = center_of_mass(footprint)
            center_x, center_y = int(center[0]), int(center[1])

            # find the boundary
            n_width = crop_strip_width
            strip_x = np.array(np.sum(footprint[:, center_y - n_width: center_y + n_width], axis=1))
            strip_y = np.flip(np.array(np.sum(footprint[center_x - n_width: center_x + n_width, :], axis=0)))
            threshold_x = 0.5 * np.max(strip_x)
            threshold_y = 0.5 * np.max(strip_y)

            left_x = np.amin(np.where(strip_x > threshold_x))
            right_x = np.amax(np.where(strip_x > threshold_x))
            up_y = np.amin(np.where(strip_y > threshold_y))
            down_y = np.amax(np.where(strip_y > threshold_y))

            center_x = coordinate_h[center_x]
            center_y = coordinate_v[self.__camera.get_configuration_file().IMAGE_SIZE_PIXEL_V - center_y]  # image is flipped vertically
            width_x = (right_x - left_x) * self.__camera.get_configuration_file().PIXEL_SIZE * 1e3
            width_y = (down_y - up_y) * self.__camera.get_configuration_file().PIXEL_SIZE * 1e3

            print("Crop Region: Center (HxV) =", round(center_x, 4), round(center_y, 4),
                  "mm, Dimension (HxV) =", round(width_x, 4), round(width_y, 4), "mm")

            kwargs["calculate_projections_over_noise"] = False
            beam_properties = super(EpicsCameraDirectBeamManager, self).get_beam_properties(beam=beam, **kwargs)

            beam_properties.set_parameter(DirectBeamImageProperties.CENTROID_H, center_x)
            beam_properties.set_parameter(DirectBeamImageProperties.CENTROID_V, center_y)
            beam_properties.set_parameter(DirectBeamImageProperties.FWHM_H,     width_x)
            beam_properties.set_parameter(DirectBeamImageProperties.FWHM_V,     width_y)

            return beam_properties

    def _get_beam_histogram(self, beam : Beam, **kwargs) -> BeamProperties:
        coordinate_h = beam.data["coordinate_h"]
        coordinate_v = beam.data["coordinate_v"]
        histogram    = beam.data["image"]

        if kwargs.get("denoise", False):
            denoise_algorithm = kwargs.get("denoise_algorithm", "simple")

            if denoise_algorithm == "simple":
                noise_mask      = kwargs.get("noise_mask", [0, 10, 0, 10])
                noise_threshold = kwargs.get("noise_threshold", 1.0)
                bad_pixels      = kwargs.get("bad_pixels", [])

                histogram, histogram_h, histogram_v = calculate_projections_over_noise(histogram=histogram,
                                                                                       noise_mask=noise_mask,
                                                                                       noise_threshold=noise_threshold,
                                                                                       bad_pixels=bad_pixels)
            else:
                raise ValueError("Denoise algorithm not recognized")
        else:
            histogram_h = histogram.sum(axis=1)
            histogram_v = histogram.sum(axis=0)

        return BeamProperties(histogram=histogram,
                              coordinate_h=coordinate_h,
                              coordinate_v=coordinate_v,
                              histogram_h=histogram_h,
                              histogram_v=histogram_v)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2025, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2025. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #
import abc
import numpy as np

from aps.beamline_driver.beam_management.facade import BeamProperties, Beam, MethodCallingType
from aps.beamline_driver.beam_management.abstract_beam_manager import DirectBeamImageBeamManager
from aps.beamline_driver.common.utility import generate_noise, calculate_projections_over_noise
from scipy.ndimage import gaussian_filter, uniform_filter

from hybrid_methods.coherence.hybrid_screen import StdIOHybridListener

try:    import bluesky.plan_stubs as bps
except: print("bluesky not installed")

EXPERIMENTAL_NOISE_TO_SIGNAL_RATIO = (100 / 5e4)
NOISE_DEFAULT_VALUE = 70 * EXPERIMENTAL_NOISE_TO_SIGNAL_RATIO

class NullHybridListener(StdIOHybridListener):
    def status_message(self, message : str): pass
    def set_progress_value(self, percentage : int): pass

class S4BeamManager(DirectBeamImageBeamManager):
    def __init__(self, method_calling_type=MethodCallingType.STANDARD, **kwargs):
        super(S4BeamManager, self).__init__(method_calling_type=method_calling_type, **kwargs)

        self._to_micron              = kwargs.get("to_micron", True)
        self._nbins_h                = kwargs.get("nbins_h", 201)
        self._nbins_v                = kwargs.get("nbins_v", 201)
        self._xrange                 = kwargs.get("xrange", None)
        self._yrange                 = kwargs.get("yrange", None)
        self._add_noise              = kwargs.get("add_noise", False)
        self._noise                  = kwargs.get("noise", NOISE_DEFAULT_VALUE if self._add_noise else None)
        self._percentage_fluctuation = kwargs.get("percentage_fluctuation", 10.0) * 1e-2
        self._calculate_over_noise   = kwargs.get("calculate_over_noise", False)
        self._noise_threshold        = kwargs.get("noise_threshold", 1.5)

        if not self._xrange is None: self._xrange = (np.array(self._xrange) * (1e-6 if self._to_micron else 1.0)).tolist()
        if not self._yrange is None: self._yrange = (np.array(self._yrange) * (1e-6 if self._to_micron else 1.0)).tolist()

    def set_method_calling_type(self, value):
        super(S4BeamManager, self).set_method_calling_type(value)

        if self.method_calling_type() == MethodCallingType.GENERATOR:
            setattr(self, "_set_initial_flux",  self.__set_initial_flux_generator)
            setattr(self, "_get_beam",          self.__get_beam_generator)
        else:
            setattr(self, "_set_initial_flux",  self.__set_initial_flux)
            setattr(self, "_get_beam",          self.__get_beam)

    def __set_initial_flux_generator(self, *args, **kwargs): yield from bps.null()

    def __set_initial_flux(self, *args, **kwargs): pass

    def __get_beam_generator(self, *args, **kwargs) -> Beam:
        yield from bps.null()
        return Beam(self._get_simulated_beam(**kwargs))

    def __get_beam(self, *args, **kwargs) -> Beam:
        return Beam(self._get_simulated_beam(**kwargs))

    @abc.abstractmethod
    def _get_simulated_beam(self) -> Beam: raise NotImplementedError

    def _get_beam_histogram(self, beam : Beam, **kwargs) -> BeamProperties:
        s4_beam = beam.data

        ticket = s4_beam.histo2(col_h=1,
                                col_v=3,
                                nbins_h=self._nbins_h,
                                nbins_v=self._nbins_v,
                                nolost=1,
                                xrange=self._xrange,
                                yrange=self._yrange)

        if not kwargs.get("blur_image", False):  histogram = ticket["histogram"]
        else:
            filter = kwargs.get("filter", "gaussian")
            if   filter == "gaussian": histogram = gaussian_filter(ticket["histogram"], sigma=kwargs.get("sigma", 2))
            elif filter == "uniform":  histogram = uniform_filter(ticket["histogram"],  size=kwargs.get("sigma", 5))

        coordinate_h = ticket['bin_h_center'] * (1e6 if self._to_micron else 1.0)
        coordinate_v = ticket['bin_v_center'] * (1e6 if self._to_micron else 1.0)

        if self._add_noise:
            histogram = generate_noise(histogram, self._noise, self._percentage_fluctuation)

            if self._calculate_over_noise:
                _, histogram_h, histogram_v = calculate_projections_over_noise(histogram, self._noise_threshold)
            else:
                histogram_h = histogram.sum(axis=1)
                histogram_v = histogram.sum(axis=0)
        else:
            histogram_h = ticket['histogram_h']
            histogram_v = ticket['histogram_v']

        if kwargs.get("scan_mode", False):
            return BeamProperties(coordinate_h=coordinate_h,
                                  coordinate_v=coordinate_v,
                                  profile_h=histogram_h,
                                  profile_v=histogram_v)
        else:
            return BeamProperties(histogram=histogram,
                                  coordinate_h=coordinate_h,
                                  coordinate_v=coordinate_v,
                                  histogram_h=histogram_h,
                                  histogram_v=histogram_v)
